@@ -261,7 +261,7 @@ handle_applications_finished_cb (EphyAboutHandler       *handler,
                             "<td class=\"data\"><div class=\"appname\">%s</div><div class=\"appurl\">%s</div></td>"
                             "<td class=\"input\"><input type=\"button\" value=\"%s\" onclick=\"deleteWebApp('%s');\"></td>"
                             "<td class=\"date\">%s <br /> %s</td></tr></tbody>",
-                            app->name, app->icon_url, app->name, app->url, _("Delete"), app->name,
+                            app->id, app->icon_url, app->name, app->url, _("Delete"), app->id,
                             /* Note for translators: this refers to the installation date. */
                             _("Installed on:"), app->install_date);
   }
