@@ -39,9 +39,8 @@
  * analyze modules. Refer to doc/debugging.txt for more information.
  */
 
-static const char *ephy_debug_break = NULL;
-
 #if DEVELOPER_MODE
+static const char *ephy_debug_break = NULL;
 static GHashTable *ephy_profilers_hash = NULL;
 static char **ephy_profile_modules;
 static gboolean ephy_profile_all_modules;
@@ -183,7 +182,7 @@ ephy_profiler_dump (EphyProfiler *profiler)
 {
   double seconds;
 
-  g_return_if_fail (profiler != NULL);
+  g_assert (profiler != NULL);
 
   seconds = g_timer_elapsed (profiler->timer, NULL);
 
@@ -195,7 +194,7 @@ ephy_profiler_dump (EphyProfiler *profiler)
 static void
 ephy_profiler_free (EphyProfiler *profiler)
 {
-  g_return_if_fail (profiler != NULL);
+  g_assert (profiler != NULL);
 
   g_timer_destroy (profiler->timer);
   g_free (profiler->name);
